"""Command-line interface."""

import logging
import os
import sys

import click
import colorlog

from quark_ctl.version import version


try:
    import qamlib

    HAVE_QAMLIB = True
except ImportError:
    HAVE_QAMLIB = False


def _logger_conf(debug: bool) -> None:
    """Setup the logging environment."""
    log = logging.getLogger()
    format_str = "%(asctime)s - %(levelname)-8s - %(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(format_str, date_format)
    if os.isatty(2):
        cformat = "%(log_color)s" + format_str
        colors = {
            "DEBUG": "blue",
            "INFO": "reset",
            "WARNING": "bold_yellow",
            "ERROR": "bold_red",
            "CRITICAL": "bold_red",
        }
        formatter = colorlog.ColoredFormatter(cformat, date_format, log_colors=colors)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    log.addHandler(stream_handler)
    if debug:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)
    return


@click.group(context_settings={"show_default": True})
@click.option("--debug", is_flag=True)
@click.option("--ip", help="IP for remote camera")
@click.option("--port", help="Port for remote camera")
@click.version_option(version)
@click.pass_context
def cli(ctx, debug, ip, port):
    """Commandline tool to interact with a qtec camera."""
    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug
    ctx.obj["IP"] = ip
    ctx.obj["PORT"] = port
    _logger_conf(debug)


#
# Exposure sequence
#
@cli.command(no_args_is_help=True)
@click.option("--exposure", help="Exposure to set")
@click.option("--device", help="Video device", default="/dev/qtec/video0")
def exposure_sequence(  # noqa C901
    exposure,
    device,
) -> None:
    """Create (and set) an exposure sequence."""
    if not HAVE_QAMLIB:
        logging.error("Qamlib not available")
        exit(1)

    cam = qamlib.Camera(device)

    if "line_exposure_sequence" not in cam.list_controls():
        logging.error("Camera does not have 'line_exposure_sequence' control")
        sys.exit(1)

    exp_seq = cam.get_ext_control("line_exposure_sequence").value
    exp_seq[:] = exposure
    exp_seq = cam.set_ext_control(
        "line_exposure_sequence", qamlib.ArrayControlValue(exp_seq)
    )


#
# Camera configuration save/load
#
@cli.group(no_args_is_help=True)
@click.pass_context
def config(ctx) -> None:
    """Save/load/reset camera configuration."""
    if not HAVE_QAMLIB:
        logging.error("qamlib not available.")
        exit(1)
    try:
        import quarklib  # noqa F401
    except ImportError:
        logging.error("quarklib not available.")
        exit(1)


__DEFAULT_CONFIG_PATH = "/var/lib/qtec/config{VIDEO_NUM}.json"


def _get_config_path(device: str, file: str):
    import re

    if file == __DEFAULT_CONFIG_PATH:
        if os.path.dirname(device) == "/dev/qtec":
            match = re.match("/dev/qtec/video([0-9]+)", device)
            if match is not None:
                return file.format(VIDEO_NUM=match.group(1))
            else:
                logging.error("Could not extract video device number from {device}")
                sys.exit(1)
        else:
            logging.error(
                "Cannot use default file path with video device that is not in "
                "/dev/qtec"
            )
            sys.exit(1)

    return file


@config.command()
@click.argument("controls", nargs=-1, type=str)
@click.option(
    "--file",
    type=click.Path(),
    default=__DEFAULT_CONFIG_PATH,
    help=(
        "File to save the configuration to. Default where VIDEO_NUM will be the"
        " number of the video device from /dev/qtec/videoX"
    ),
)
@click.option(
    "--device",
    help="Device to save configuration of",
    default="/dev/qtec/video0",
)
@click.option(
    "--no-controls",
    help="Don't save any controls in configuration",
    is_flag=True,
)
@click.option("--no-crop", help="Don't save croppings in configuration", is_flag=True)
@click.option("--no-format", help="Don't save format in configuration", is_flag=True)
@click.option(
    "--no-framerate", help="Don't save framerate in configuration", is_flag=True
)
@click.option(
    "--include-protected-controls",
    help=(
        "ADC gain, black level or VRAMP are sensor specific and should almost "
        "never be touched. Save their values anyway."
    ),
    is_flag=True,
)
def save(
    file,
    device,
    controls,
    no_controls: bool,
    no_crop: bool,
    no_format: bool,
    no_framerate: bool,
    include_protected_controls: bool,
) -> None:
    """Saves camera configuration to a .json file.

    If no CONTROLS are specified all controls (except protected ones) are saved
    """
    import quarklib

    if os.path.isdir(file):
        logging.error(f"Path is a directory: {file}")
        sys.exit(1)

    cam = qamlib.Camera(device)

    if len(controls) == 0:
        logging.debug("No controls selected so defaulting to all")
        controls = None

    if no_controls:
        logging.debug("No controls chosen")
        controls = []

    file = _get_config_path(device, file)

    logging.info(f"Saving configuration to: {file}")
    quarklib.config.save(
        cam,
        path=file,
        controls=controls,
        format=(not no_format),
        crop=(not no_crop),
        framerate=(not no_framerate),
        protect_controls=(not include_protected_controls),
    )


@config.command()
@click.option(
    "--file",
    type=click.Path(),
    default=__DEFAULT_CONFIG_PATH,
    help=(
        "File to save the configuration to. Default where VIDEO_NUM will be the"
        " number of the video device from /dev/qtec/videoX"
    ),
)
@click.option(
    "--device",
    help="Device to save configuration of",
    default="/dev/qtec/video0",
)
@click.argument("controls", nargs=-1, type=str)
@click.option("--no-controls", help="Ignore controls from configuration", is_flag=True)
@click.option("--no-crop", help="Ignore crop from configuration", is_flag=True)
@click.option("--no-format", help="Ignore format from configuration", is_flag=True)
@click.option(
    "--no-framerate", help="Ignore framerate from configuration", is_flag=True
)
@click.option(
    "--include-protected-controls",
    help=(
        "ADC gain, black level or VRAMP are sensor specific and should almost "
        "never be touched. Load their values anyway."
    ),
    is_flag=True,
)
def load(
    file,
    device,
    controls,
    no_controls,
    no_crop,
    no_format,
    no_framerate,
    include_protected_controls,
) -> None:
    """Loads a camera configuration.

    This will fail if the camera is streaming.

    If no CONTROLS are specified all controls (except protected ones) are saved
    """
    import quarklib

    file = _get_config_path(device, file)

    if not os.path.isfile(file):
        logging.warning(f"No such file: {file}")
        exit(1)

    cam = qamlib.Camera(device)

    if len(controls) == 0:
        logging.debug("No controls selected so defaulting to all")
        controls = None

    if no_controls:
        logging.debug("No controls chosen")
        controls = []

    logging.info(f"Loading configuration from: {file}")
    quarklib.config.load(
        cam,
        path=file,
        controls=controls,
        format=(not no_format),
        crop=(not no_crop),
        framerate=(not no_framerate),
        protect_controls=(not include_protected_controls),
    )


@config.command()
@click.option(
    "--device",
    help="Device of which to reset the controls",
    default="/dev/qtec/video0",
)
@click.option(
    "--include-protected-controls",
    help=(
        "ADC gain, black level or VRAMP are sensor specific and should almost "
        "never be touched. Reset their values anyway."
    ),
    is_flag=True,
)
@click.argument("controls", nargs=-1, type=str)
def reset_controls(device, controls, include_protected_controls) -> None:
    """Reset controls of a camera to their default values.

    If no CONTROLS are specified all controls (except protected ones) are saved.
    """
    import quarklib

    cam = qamlib.Camera(device)

    if len(controls) == 0:
        logging.debug("No controls selected so defaulting to all")
        controls = None

    quarklib.config.reset_controls(cam, controls, include_protected_controls)


@config.command()
def load_default_configs():
    """Load configuration for all heads.

    This will try to load the configuration for each head from /var/lib/qtec.
    NOTE: This function is used for the quark-init-load service to load
    configuration at boot.
    """
    import re

    import quarklib

    regex = re.compile("video([0-9]+)")

    for dev in os.listdir("/dev/qtec/"):
        match = regex.match(dev)
        if match is not None:
            device = f"/dev/qtec/video{match.group(1)}"
            cam = qamlib.Camera(device)

            file = __DEFAULT_CONFIG_PATH.format(VIDEO_NUM=match.group(1))
            logging.info(f"Loading configuration from: {file}")

            if not os.path.isfile(file):
                logging.warning(f"No such file: {file}")
                continue

            try:
                quarklib.config.load(cam, path=file)
            except Exception:
                logging.warning(f"Failed to load config {file} to {device}")


#
# Camera auto-exposure
#
@cli.command()
@click.option(
    "--device",
    help="Device of which to reset the controls",
    default="/dev/qtec/video0",
)
@click.option(
    "--crop",
    nargs=4,
    type=int,
    help=(
        "LEFT TOP WIDTH HEIGHT - "
        "crop of current frame output to look at for auto exposure"
    ),
)
@click.option(
    "--max-iterations",
    type=int,
    help="Maximum number of frames to use for auto exposure",
    default=200,
)
@click.option(
    "--target-value",
    type=int,
    help="Specify a target value for the saturation value",
)
def auto_exposure(device, crop, max_iterations, target_value) -> None:
    """One-shot auto-exposure."""
    if not HAVE_QAMLIB:
        logging.error("qamlib not available.")
        exit(1)

    try:
        from quarklib.isp import SimpleAutoExposure
        from quarklib.isp import one_shot_auto_exposure
    except ImportError:
        logging.error("quarklib AutoExposure not available")
        exit(1)

    cam = qamlib.Camera(device)

    width, height = cam.get_resolution()

    if crop is not None:
        cl, ct, cw, ch = crop
        if cl + cw > width or ct + ch > height:
            print("The specified crop does not fit in current frame size")
            exit(1)

        if cl < 0 or ct < 0:
            print("Crop left or top cannot be negative")
            exit(1)

        if cw < 1 or ch < 1:
            print("Crop width or height cannot be less than 1")
            exit(1)

    auto_exp = SimpleAutoExposure(cam, crop=crop, target_value=target_value)

    one_shot_auto_exposure(
        cam,
        auto_exposure=auto_exp,
        max_iterations=max_iterations,
    )


@cli.command()
@click.option("--output", default=None, type=str, help="Output histogram to .csv file")
@click.option(
    "--bins",
    default=None,
    type=int,
    help="Specify number of bins (uses automatic bin count if omitted)",
)
@click.option(
    "--crop",
    default=None,
    type=int,
    nargs=4,
    help="Specify ROI (top, left, width, height)",
)
@click.option("--density", is_flag=True, help="Use pdf or counts")
@click.option(
    "--grey-convert",
    is_flag=True,
    help="Convert the image to grayscale before processing.",
)
@click.option(
    "--n", default=1, help="Average stats over n frames (not used in live mode)"
)
@click.option(
    "--alpha",
    type=float,
    default=None,
    help="Alpha for exponential decaying average (only used in live mode)",
)
@click.option("--live", is_flag=True, help="Use live mode.")
@click.option("--device", type=str, default=None, help="Optional camera device path.")
def stats(output, bins, crop, density, grey_convert, n, alpha, live, device):
    """View and export camera statistics (including histograms)."""
    from quark_ctl._util import filter_missing
    from quark_ctl._util import print_missing_dependency_info

    missing = filter_missing(["textual", "rich", "plotille", "qamlib", "pandas"])
    if len(missing) > 0:
        print_missing_dependency_info(missing)
        sys.exit(1)

    from quark_ctl import stats

    if live:
        stats.stats_live(output, bins, crop, density, grey_convert, alpha, device)
    else:
        stats.single_shot(n, output, bins, crop, density, grey_convert, device)


@cli.command(no_args_is_help=True)
@click.option(
    "--iterations",
    default=1,
    type=click.IntRange(1),
    help="Number of frames",
)
@click.option("--avg", is_flag=True, help="Average of iterations")
@click.option("--device", help="Video device", default="/dev/qtec/video0")
@click.option("--demosaic", is_flag=True, help="Apply demosaicing to bayer images")
@click.option("--makedir", is_flag=True, help="Create new directory")
@click.argument("image_name", type=str)
def save_image(image_name, iterations, avg, device, demosaic, makedir):
    """Save single/multiple/average frame(s).

    NOTE: This feature supports: PNG, PNM and PAM file formats.
    """
    if not HAVE_QAMLIB:
        logging.error("Qamlib not available")
        exit(1)
    from quark_ctl.save_img import save_image_core

    save_image_core(image_name, iterations, avg, device, demosaic, makedir)


if __name__ == "__main__":
    cli()
