"""Save image functions."""

import logging
import math
import os
import sys

import numpy as np
import qamlib


try:
    import quarklib

    HAVE_QUARKLIB = True
except ImportError:
    HAVE_QUARKLIB = False

try:
    import cv2

    HAVE_OPENCV = True
except ImportError:
    HAVE_OPENCV = False


# tuple of bayer formats of color (BA81 is BGGR in V4L UI)
BAYER_FORMATS = ("RGGB", "GRBG", "GBRG", "BA81")

# tuple of color formats
COLOR_IMAGE_FORMATS = ("RGB3", "RGB4", "RGB6", "BGR3", "BGR4")

# supported file formats
SUPPORTED_FILE_FORMATS = (".pam", ".png", ".pnm")


def save_image_core(
    image_name: str,
    iterations: int,
    avg: bool,
    device: str,
    demosaic: bool,
    makedir: bool,
):
    """Save and optionally average a series of camera frames.

    :param image_name: Desired name for the saved image.
    :param iterations: Amount of frames to be taken.
    :param avg: The averaged image of the amount of iterations.
    :param device: Camera device.
    :param demosaic: A flag for determining if the frame should be demosaiced
        (if bayer format).
    """
    directory_path, extension = os.path.splitext(image_name)
    dir_name = os.path.dirname(directory_path)

    if not os.path.exists(dir_name) and dir_name != "":
        if makedir:
            os.makedirs(dir_name)
            print(f"Directory path {dir_name!r} has been created.")
            logging.debug(f"Directory path {dir_name!r} has been created.")
        else:
            logging.debug(f"Directory {dir_name!r} does NOT exist.")
            raise FileNotFoundError(f"Directory {dir_name!r} does NOT exist.")

    elif os.path.exists(dir_name) and makedir:
        logging.debug(
            f"Directory path {dir_name!r} already exists. Try without --makedir option"
        )
        raise FileExistsError(
            f"Directory path {dir_name!r} already exists. Try without --makedir option"
        )

    # .lower() is used to mitigate human error
    extension = extension.lower()
    if not HAVE_QUARKLIB:
        raise ImportError("Quarklib not available")

    elif extension == ".png" or demosaic:
        if not HAVE_OPENCV:
            raise ImportError(
                "OpenCV is missing. To install OpenCV, use the "
                "following command:\n\napt-get install python3-opencv\n"
            )

    if extension in SUPPORTED_FILE_FORMATS:
        process_frames(
            image_name, iterations, avg, directory_path, extension, demosaic, device
        )

    else:
        raise ValueError(
            f"Unsupported image type: {extension}\n"
            f"Supported image types: {SUPPORTED_FILE_FORMATS}"
        )


def process_frames(
    image_name: str,
    iterations: int,
    avg: bool,
    directory_path: str,
    extension: str,
    demosaic: bool,
    device: qamlib.Camera,
):
    """Starts frame capturing process.

    :param image_name: The full name to save to.
    :param iterations: The number of frames to be taken.
    :param avg: Average of the amount of frames.
    :param directory_path: Directory path for frames to be saved.
    :param extension: Image file format.
    :param demosaic: A flag for determining if the frame should be demosaiced
        (if bayer format).
    """
    cam = qamlib.Camera(device)

    try:
        with cam:
            img_format = cam.get_format().pixelformat.fourcc
            # multiple frames and/or average
            if iterations > 1:
                multiple_frames(
                    cam,
                    img_format,
                    iterations,
                    avg,
                    demosaic,
                    directory_path,
                    extension,
                )
            # single frame
            else:
                _, frame = cam.get_frame()

                execute_save_function(
                    image_name, frame, img_format, demosaic, extension
                )
    except qamlib.V4L2Exception as e:
        raise qamlib.V4L2Exception(f"Error getting frame from camera: {e}") from e


def needs_byteswap(frame: np.ndarray) -> bool:
    """Check if a byte-swap is needed for the PAM format.

    :param frame: The frame to save.
    """
    # 8-bit
    if frame.itemsize == 1:
        return False

    # Big Endian
    if frame.dtype.byteorder == ">":
        return False

    # Big Endian native byte order
    if frame.dtype.byteorder == "=" and sys.byteorder == "big":
        return False

    return True


def execute_save_function(
    name: str, frame: np.ndarray, img_format: str, demosaic: bool, extension: str
) -> None:
    """Save file format function.

    :param name: The full name to save to.
    :param frame: The frame to save.
    :param img_format: The image/frame format fourcc (RGB3, BGR3, BA81, ...).
    :param demosaic: A flag for determining if the frame should be demosaiced
        (if bayer format).
    :param extension: Image file format.
    """
    if demosaic:
        frame = process_demosaiced_img(img_format, frame)
    if extension == ".png":
        save_png(name, frame, img_format)
    elif extension == ".pam":
        save_pam(name, frame, img_format)
    elif extension == ".pnm":
        save_pnm_wrapper(name, frame, img_format)


def save_pam(name: str, frame: np.ndarray, img_format: str) -> None:
    """Save frame as PAM file.

    NOTE: For 16 bit frames PAM uses big endian, so we may have to do a
    byte-swap, this can decrease performance, but it is possible to improve the
    byte-swap speed by doing it in-place ``frame =
    frame.byteswap(True).newbyteorder()`` before calling :func:`save_pam`

    :param name: The full name to save to.
    :param frame: The frame to save.
    :param img_format: The image/frame format fourcc (RGB3, BGR3, BA81, ...).
    """
    with open(name, "wb") as f:
        height, width, channels = frame.shape
        # get the format: RGB, Bayer, etc.
        type = "P7"

        if frame.nbytes == width * height * channels:
            if frame.dtype == "uint8":
                max_val = 255
            else:
                max_val = 65535
        else:
            max_val = 65535

        header_dict = {
            "WIDTH": width,
            "HEIGHT": height,
            "DEPTH": channels,
            "MAXVAL": max_val,
            "TUPLTYPE": img_format,
        }

        # Create the header
        pam_header = f"{type}\n"
        for key, value in header_dict.items():
            pam_header += f"{key} {value}\n"
        pam_header += "ENDHDR\n"

        f.write(bytearray(pam_header, "ascii"))

        if needs_byteswap(frame):
            frame.byteswap().tofile(f)
        else:
            frame.tofile(f)


def save_png(name: str, frame: np.ndarray, img_format: str) -> None:
    """Save frame as PNG file.

    :param name: The full name to save to.
    :param frame: The frame to save.
    :param img_format: The image/frame format fourcc (RGB3, BGR3, BA81, ...).
    :param demosaic: A flag for determining if the frame should be demosaic
        (if bayer format).
    """
    if img_format in COLOR_IMAGE_FORMATS:
        frame = rgb_to_bgr(img_format, frame)

    if not cv2.imwrite(name, frame):
        raise Exception(f"Failed to write PNG {name}")


def save_pnm_wrapper(name: str, frame: np.ndarray, img_format: str) -> None:
    """Wrapper for saving pnm. img_format is needed for decision making.

    :param name: The full name to save to.
    :param frame: The frame to save.
    :param img_format: The image/frame format fourcc (RGB, BGR, BA81, ...).
    """
    match img_format:
        case img_format if img_format in BAYER_FORMATS:
            # frame has to be reversed from BGR -> RGB if image is demosaiced.
            frame = frame[:, :, ::-1].copy()
            quarklib.image.save_pnm(name, frame)
        case "BGR3":
            frame = frame[:, :, ::-1].copy()
            quarklib.image.save_pnm(name, frame)
        case "GREY" | "RGB3" | "Y16" | "RGB6":
            quarklib.image.save_pnm(name, frame)
        case _:
            raise ValueError(
                f"Incorrect image format {img_format} for pnm."
                f"Supported: RGB3, BGR3, GREY, Y16"
            )


def process_demosaiced_img(img_format: str, frame: np.ndarray):
    """Debayers/Demosaics bayer formats.

    :param img_format: The image/frame format (RGB, BGR, BA81, ...).
    :param frame: The frame to save.
    """
    match img_format:
        case "GBRG":
            return cv2.cvtColor(frame, cv2.COLOR_BayerGR2BGR)
        case "GRBG":
            return cv2.cvtColor(frame, cv2.COLOR_BayerGB2BGR)
        case "RGGB":
            return cv2.cvtColor(frame, cv2.COLOR_BayerBG2BGR)
        case "BA81":
            return cv2.cvtColor(frame, cv2.COLOR_BayerRG2BGR)
        case _:
            raise ValueError(
                f"Invalid fourcc: {img_format}. Supported: {BAYER_FORMATS}"
            )


def rgb_to_bgr(img_format: str, frame: np.ndarray):
    """Converts to BGR if img_format is RGB.

    :param img_format: The image/frame format fourcc (RGB, BGR, BA81, ...).
    :param frame: The frame to save.
    :return: Returns the frame as BGR3.
    """
    if img_format not in ["RGB3", "BGR3", "RGB6"]:
        raise ValueError(
            f"Unsupported file format {img_format}. \nSupported formats RGB3, BGR3, RGB6"
        )

    match img_format:
        case "RGB3" | "RGB6":
            return cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        case "BGR3":
            return frame
        case _:
            raise ValueError(
                f"Invalid fourcc: {img_format}. Supported: RGB3, RGB6, BGR3"
            )


def multiple_frames(
    cam: object,
    img_format,
    iterations: int,
    avg: bool,
    demosaic: bool,
    directory_path: str,
    extension: str,
):
    """Captures multiple frames.

    Captures multiple frames and saves them either individually or
    as a single averaged frame.

    :param cam: Camera object
    :param img_format: The image/frame format fourcc (RGB, BGR, BA81, ...).
    :param iterations: Amount of frames to be taken.
    :param avg: The averaged image of the amount of iterations.
    :param demosaic: A flag for determining if the frame should be demosaiced.
    :param directory_path: Directory path for frames to be saved.
    :param extension: Image file format.
    """
    cumulative_sum = None
    digits = int(math.log10(iterations)) + 1

    for im in range(iterations):
        _, frame = cam.get_frame()

        if avg:
            frame = frame.astype(np.float64)
            if cumulative_sum is None:
                cumulative_sum = frame
            else:
                cumulative_sum += frame
        else:
            execute_save_function(
                f"{directory_path}{im + 1:0{digits}}{extension}",
                frame,
                img_format,
                demosaic,
                extension,
            )
    # calculate average
    if avg:
        average_frame = (cumulative_sum / iterations).astype(np.uint8)

        execute_save_function(
            directory_path + extension,
            average_frame,
            img_format,
            demosaic,
            extension,
        )
