"""Camera frame statistics."""

from dataclasses import dataclass
from typing import List
from typing import Tuple


try:
    import qamlib

    HAVE_QAMLIB = True
except ImportError:
    HAVE_QAMLIB = False

import logging
import time
from threading import Thread

import numpy as np
import pandas as pd
import plotille as plt
from rich.console import Console
from rich.table import Table
from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Grid
from textual.events import Resize
from textual.logging import TextualHandler
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import DataTable
from textual.widgets import Footer
from textual.widgets import Input
from textual.widgets import Label
from textual.widgets import Static
from textual.widgets import Switch


def stats_live(
    output: str | None,
    bins: int | None,
    crop: Tuple[int, int, int, int] | None,
    density: bool,
    to_gray: bool,
    alpha: float | None,
    device: str | None = None,
):
    """Show camera statistics live in the terminal.

    :param output: Optional output path. Should not contain a file
            extension (this is added by the script).
    :param bins: Number of histogram bins. Determined automatically
             by the `np.histogram` function if omitted.
    :param crop: Use an ROI for processing.
    :param density: Output histogram values as densities instead of counts.
            Default is False.
    :param to_gray: Convert the image to gray-scale before processing.
            Default is False.
    :param alpha: Alpha value for exponentially decaying average.
    :param device: Optional camera device path.

    """
    cam = qamlib.Camera(device)
    with cam:
        app = _StatsApp(cam, output, bins, crop, density, to_gray, alpha)
        app.run()


def single_shot(
    n: int,
    output: str | None,
    bins: int | None = None,
    crop: Tuple[int, int, int, int] | None = None,
    density: bool = False,
    to_gray: bool = False,
    device: str | None = None,
):
    """Print camera statistics to stdout and (optionally) to csv.

    :param n: Average over n samples.
    :param output: Optional output path. Should not contain a file
            extension (this is added by the script).
    :param bins: Number of histogram bins. Determined automatically
             by the `np.histogram` function if omitted.
    :param crop: Use an ROI for processing.
    :param density: Output histogram values as densities instead of counts.
            Default is False.
    :param to_gray: Convert the image to gray-scale before processing.
            Default is False.
    :param device: Optional camera device path.

    """
    cam = qamlib.Camera(device)
    with cam:
        pixel_format = _setup_camera(cam, crop)

        average_frame = _capture_data(cam, n)

    logging.debug("camera stopped")

    average_frame = _to_gray(average_frame, pixel_format, to_gray)

    df = _get_stats(average_frame, pixel_format)
    hist_vals, bin_edges = _get_histogram(average_frame, bins, density, pixel_format)

    _print_to_console(bin_edges, to_gray, df, pixel_format, hist_vals)

    if output is not None:
        _write_output(bin_edges, df, hist_vals, output, pixel_format)
        logging.info(f"wrote result to {output}_hist.csv and {output}_stats.csv")

    _reset_crop(cam, crop)


class _StatsApp(App):
    DEFAULT_CSS = """
    Screen {
        background: $background;
    }
    """

    BINDINGS = [
        Binding(key="escape, q", action="quit", description="Quit"),
        Binding(
            key="s",
            action="save",
            description="Save output",
        ),
        Binding(
            key="p",
            action="settings",
            description="Settings",
        ),
        Binding(
            key="k, up",
            action="scroll_up",
            description="Scroll up",
            priority=True,
        ),
        Binding(
            key="j, down",
            action="scroll_down",
            description="Scroll down",
            priority=True,
        ),
    ]

    def cam_loop(self):
        self.frame = None

        try:
            while not self.should_exit:
                time.sleep(0.005)

                frame = _capture_data(self.cam, 1)

                if self.to_gray:
                    frame = np.mean(frame, axis=0, keepdims=True)

                if self.alpha and self.frame is not None:
                    self.frame = self.frame * (self.alpha - 1) + frame * self.alpha
                else:
                    self.frame = frame

                df = _get_stats(frame, self.pixel_format)
                self.df = df
                self.update_stats(df)
                self.update_histograms(frame)

        except Exception as e:
            self.panic(f"Error: {e}")

    def __init__(
        self,
        cam: qamlib.Camera,
        output: str | None,
        bins: int | None,
        crop: Tuple[int, int, int, int] | None,
        density: bool,
        to_gray: bool,
        alpha: float | None,
    ):
        self.output = output
        self.bins = bins
        self.crop = crop
        self.density = density
        self.to_gray = to_gray
        self.alpha = alpha

        self.should_exit = False

        if self.output is None:
            self.output = "output"

        log = logging.getLogger()
        log.handlers.clear()
        handler = TextualHandler()
        log.addHandler(handler)

        self.cam = cam
        self.pixel_format = _setup_camera(cam, crop)

        if self.to_gray and self.pixel_format.n_channels() > 1:
            self.pixel_format.channels = ["Y"]

        super().__init__()

    def on_mount(self):
        table = self.query_one(DataTable)
        self.histogram = self.query_one(_HistogramContainer)
        self.table = table

        zero_fill = [0 for _ in range(8)]

        table.add_columns(
            "Channel",
            "Mean",
            "Std",
            "Min",
            "Max",
            "Min (count)",
            "Max (count)",
            "Median",
        )
        for _ in range(self.pixel_format.n_channels()):
            table.add_row(*zero_fill)

        self.worker = Thread(target=self.cam_loop)
        self.worker.start()

    def update_stats(self, df):
        for y, r in df.iterrows():
            for x, cell in enumerate(r):
                if isinstance(cell, str) or isinstance(cell, int):
                    val = f"{cell}"
                else:
                    val = f"{cell:.2f}"
                self.table.update_cell_at((y, x), val, update_width=True)

    def update_histograms(self, frame):
        self.histogram.display(frame)

    def compose(self) -> ComposeResult:
        yield DataTable()
        yield _HistogramContainer(self.pixel_format, self.bins, self.density)
        yield Footer()

    def action_save(self):
        hist_vals, bin_edges = _get_histogram(
            self.frame, self.bins, self.density, self.pixel_format
        )

        _write_output(bin_edges, self.df, hist_vals, self.output, self.pixel_format)

        self.notify(
            f"Saved results at {self.output}_stats.csv and {self.output}_hist.csv",
            title="Success",
        )

    def action_settings(self):
        self.push_screen(_Settings())

    def action_quit(self):
        self.should_exit = True
        self.action_save()
        self.worker.join()
        _reset_crop(self.cam, self.crop)
        self.exit()

    def action_scroll_up(self):
        self.screen.scroll_up()

    def action_scroll_down(self):
        self.screen.scroll_down()


class _Settings(Screen):
    DEFAULT_CSS = """
    _Settings {
        align: center middle;
    }
    #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 3;
        padding: 0 1;
        width: 60;
        height: auto;
        border: thick $background 80%;
    }

    Button {
        width: 100%;
    }
    """

    BINDINGS = [
        Binding(key="enter", action="save"),
        Binding(key="escape, q", action="cancel", priority=True),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bins = self.app.bins
        self.output = self.app.output
        self.density = self.app.density

    def compose(self):
        yield Grid(
            Label("Bin count"),
            Input(str(self.bins), id="bins"),
            Label("Density"),
            Switch(self.density),
            Label("Output"),
            Input(str(self.output), id="output"),
            Button("Cancel", variant="error", id="cancel"),
            Button("Accept", variant="primary", id="accept"),
            id="dialog",
        )

    def on_input_changed(self, event: Input.Changed):
        if event.input.id == "bins":
            try:
                self.bins = int(event.value)
            except ValueError:
                self.bins = None
        elif event.input.id == "output":
            self.output = event.value

    def on_switch_changed(self, event: Switch.Changed):
        self.density = event.value

    def on_button_pressed(self, event: Button.Pressed):
        if event.button.id == "cancel":
            self.action_cancel()
        elif event.button.id == "accept":
            self.action_save()

    def action_cancel(self):
        self.app.pop_screen()

    def action_save(self):
        self.app.bins = self.bins
        self.app.histogram.bins = self.bins
        self.app.density = self.density
        self.app.histogram.density = self.density
        self.app.output = self.output
        self.app.pop_screen()


class _Histogram(Widget):
    DEFAULT_CSS = """
    .title {
        align-horizontal: center;
        text-align: center;
        width: 30%;
        height: 1;
    }

    Static {
        height: auto;
    }
    """

    def compose(self):
        # yield Label(self.title, classes="title")
        yield Static(classes="histogram", id=f"{self.title}_static")

    def __init__(self, title, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = title
        self.title = title

    def set(self, inner):
        self.log.debug(f"updating {self.title}")
        self.query_one(f"#{self.title}_static").update(inner)


class _HistogramContainer(Widget):
    DEFAULT_CSS = """
    _HistogramContainer {
        layout: grid;
        grid-size: 3;
        grid-rows: auto;
        height: auto;
        width: 100%;

    }

    _Histogram {
        layout: vertical;
        border: solid white;
        width: 100%;
        min-width: 60;
        height: auto;
    }
    """

    def __init__(
        self, pixel_format: "PixelFormat", bins: int, density: bool, *args, **kwargs
    ):
        self.pixel_format = pixel_format
        self.bins = bins

        self.density = density
        super().__init__(*args, **kwargs)

    def on_resize(self, event: Resize):
        self.styles.grid_size_columns = event.size.width // 70

    def on_mount(self):
        self.views = [self.query_one(f"#{c}") for c in self.pixel_format.channels]

    def display(self, frame=None):
        try:
            vals, bin_edges = _get_histogram(
                frame, self.bins, self.density, self.pixel_format
            )

            spacing = 50 if self.density else 42
            for c, v, view in zip(
                self.pixel_format.channels, vals, self.views, strict=False
            ):
                width = view.window_region.width - spacing
                fig = plt.hist_aggregated(v, bin_edges, width=width)

                match c:
                    case "R":
                        color = "red"
                    case "G":
                        color = "rgb(46, 139, 87)"
                    case "B":
                        color = "rgb(30, 144, 255)"
                    case _:
                        color = "white"

                view.styles.border = ("round", color)

                view.set(fig)

        except TypeError as e:
            self.log(e)

    def compose(self):
        # with TabbedContent(*self.pixel_format.channels):
        for c in self.pixel_format.channels:
            # yield Static(id=str(c), classes="histogram")
            yield _Histogram(c, id=c)


def _reset_crop(cam, crop):
    if crop is not None:
        logging.debug("crop reset")
        cam.set_crop(cam.get_crop_bounds())


def _to_gray(average_frame, pixel_format, to_gray):
    if pixel_format.n_channels() > 1 and to_gray:
        average_frame = np.mean(average_frame, axis=0, keepdims=True)
        pixel_format.channels = ["Y"]
    return average_frame


def _setup_camera(cam, crop):
    if crop is not None:
        logging.debug("camera crop set")
        cam.set_crop(*crop)
    logging.debug("camera streaming started")
    base_fmt = cam.get_format()
    pixel_format = PixelFormat.from_fourcc(base_fmt.pixelformat.fourcc)
    return pixel_format


def _dataframe_to_table(df: pd.DataFrame) -> "Table":
    table = Table()
    for c in df.columns:
        table.add_column(c)

    for _, r in df.iterrows():
        table.add_row(*[str(cell) for cell in r])

    return table


def _get_stats(channels: np.ndarray, format: "PixelFormat") -> pd.DataFrame:
    data = [
        [
            f,
            c.mean(),
            c.std(),
            c.min(),
            c.max(),
            (c == c.min()).sum(),
            (c == c.max()).sum(),
            np.median(c),
        ]
        for f, c in zip(format.channels, channels, strict=True)
    ]

    return pd.DataFrame(
        data,
        columns=[
            "Channel",
            "Mean",
            "Std",
            "Min",
            "Max",
            "Min (count)",
            "Max (count)",
            "Median",
        ],
    )


def _get_histogram(
    channels: np.ndarray, bins: int | None, density: bool, format: "PixelFormat"
) -> Tuple[np.ndarray, np.ndarray]:
    if bins is None:
        logging.debug("calculating number of bins")
        bins = np.histogram_bin_edges(channels)

    hist_vals = []
    bin_edges = None

    for c in channels:
        counts, bin_edges = np.histogram(
            c, bins, range=(0, format.max_value()), density=density
        )
        hist_vals.append(counts)

    return np.stack(hist_vals), bin_edges


def _show_table(df, console):
    table = _dataframe_to_table(df)

    console.print("\n[bold]Statistics")
    console.print(table)


def _capture_data(cam, n):
    logging.debug("capturing frames")
    _, avg = cam.get_frame()
    avg = np.float64(avg)

    for _ in range(n - 1):
        time.sleep(0.05)
        meta, frame = cam.get_frame()
        avg += np.float64(frame)

    logging.debug("capturing frames done")
    avg = avg / n

    return np.transpose(avg, axes=[2, 0, 1])


def _print_to_console(bin_edges, to_gray, df, pixel_format, hist_vals):
    console = Console()
    console.print(
        f"[bold]Format{' (converted)' if to_gray else ''}: "
        f"[{' '.join(pixel_format.channels)}] {pixel_format.bits}-bit"
    )
    _show_table(df, console)
    console.print("\n[bold]Histograms")
    for counts, c in zip(hist_vals, pixel_format.channels, strict=True):
        console.print(f"[italic]Histogram for channel {c!r}")
        fig = plt.hist_aggregated(counts, bin_edges)
        print(fig)


def _write_output(bin_edges, df, hist_vals, output, pixel_format):
    df.to_csv(f"{output}_stats.csv", index=False)
    bin_lower = bin_edges[:-1]
    data = np.vstack((bin_lower, hist_vals))
    df = pd.DataFrame(data.T, columns=["Bin start", *pixel_format.channels])
    df.to_csv(f"{output}_hist.csv", index=False)


@dataclass
class PixelFormat:
    """PixelFormat helper class."""

    fourcc: str
    channels: List[str]
    bits: int

    @staticmethod
    def from_fourcc(fourcc: str):
        """Parse from V4L2 fourcc code.

        :param fourcc: V4L2 fourcc code.
        :param fourcc: str:
        :return: The pixel format

        """
        match fourcc:
            case "GREY":
                return PixelFormat(fourcc, ["Y"], 8)
            case "Y16":
                return PixelFormat(fourcc, ["Y"], 16)
            case "QG08":
                return PixelFormat(fourcc, ["G"], 8)
            case "QG16":
                return PixelFormat(fourcc, ["G"], 16)
            case "RGB3":
                return PixelFormat(fourcc, ["R", "G", "B"], 8)
            case "RGB6":
                return PixelFormat(fourcc, ["R", "G", "B"], 16)
            case "RGB4":
                return PixelFormat(fourcc, ["X", "R", "G", "B"], 8)
            case "BGR3":
                return PixelFormat(fourcc, ["B", "G", "R"], 8)
            case "BGR4":
                return PixelFormat(fourcc, ["X", "B", "G", "R"], 8)
            case "GBRG":
                return PixelFormat(fourcc, ["GBRG"], 8)
            case "GRBG":
                return PixelFormat(fourcc, ["GRBG"], 8)
            case "RGGB":
                return PixelFormat(fourcc, ["RGGB"], 8)
            case "BA81":
                return PixelFormat(fourcc, ["BA81"], 8)

    def n_channels(self):
        """Get the number of channels.

        :returns: The number of channels.

        """
        return len(self.channels)

    def max_value(self):
        """The maximum allowed pixel-value for the format.

        :returns: The maximum allowed pixel-value.

        """
        return 2**self.bits
